/*************************************************************************
 *    CompuCell - A software framework for multimodel simulations of     *
 * biocomplexity problems Copyright (C) 2003 University of Notre Dame,   *
 *                             Indiana                                   *
 *                                                                       *
 * This program is free software; IF YOU AGREE TO CITE USE OF CompuCell  *
 *  IN ALL RELATED RESEARCH PUBLICATIONS according to the terms of the   *
 *  CompuCell GNU General Public License RIDER you can redistribute it   *
 * and/or modify it under the terms of the GNU General Public License as *
 *  published by the Free Software Foundation; either version 2 of the   *
 *         License, or (at your option) any later version.               *
 *                                                                       *
 * This program is distributed in the hope that it will be useful, but   *
 *      WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
 *             General Public License for more details.                  *
 *                                                                       *
 *  You should have received a copy of the GNU General Public License    *
 *     along with this program; if not, write to the Free Software       *
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        *
 *************************************************************************/


#include "Boundary.h"
#include "BoundaryFactory.h"
#include "BoundaryStrategy.h"
#include "AlgorithmFactory.h"
#include "Algorithm.h"
#include <CompuCell3D/Field3D/Field3DImpl.h>
#include <CompuCell3D/Field3D/Dim3D.h>
#include <CompuCell3D/Field3D/Point3D.h>
#include <CompuCell3D/Field3D/Neighbor.h>
#include <CompuCell3D/Field3D/NeighborFinder.h>
#include <map>

using namespace std;
using namespace CompuCell3D;

// Singleton
BoundaryStrategy* BoundaryStrategy::singleton;


// Default constructor
BoundaryStrategy::BoundaryStrategy(){
   
   strategy_x = BoundaryFactory::createBoundary(BoundaryFactory::no_flux);
   strategy_y = BoundaryFactory::createBoundary(BoundaryFactory::no_flux);
   strategy_z = BoundaryFactory::createBoundary(BoundaryFactory::no_flux);
   algorithm = AlgorithmFactory::createAlgorithm(AlgorithmFactory::Default,0,0,"None");

   regular = true; 
   neighborListsInitializedFlag=false;
   
}


// Constructor
BoundaryStrategy::BoundaryStrategy(string boundary_x, string boundary_y,
                   string boundary_z, string alg, int index, int size, string inputfile)

{

   strategy_x = BoundaryFactory::createBoundary(boundary_x);
   strategy_y = BoundaryFactory::createBoundary(boundary_y);
   strategy_z = BoundaryFactory::createBoundary(boundary_z);
   algorithm = AlgorithmFactory::createAlgorithm(alg, index, size, inputfile);
   regular = true;
   neighborListsInitializedFlag=false;
   
   


}

// Destructor
BoundaryStrategy::~BoundaryStrategy() {

    delete strategy_x;
    delete strategy_y;
    delete strategy_z;

}

/*
 * Check to see if the given point lies inside the field dimensions
 *
 * @param pt Point3D
 *
 * @return bool 
 */
bool BoundaryStrategy::isValid(const Point3D &pt) const {
    
    // check to see if the point lies in the dimensions before applying the
    // shape algorithm

//  return (0 <= pt.x && pt.x < dim.x &&
//                    0 <= pt.y && pt.y < dim.y &&
//                    0 <= pt.z && pt.z < dim.z );

 if (0 <= pt.x && pt.x < dim.x &&
                   0 <= pt.y && pt.y < dim.y &&
                   0 <= pt.z && pt.z < dim.z ) {
           
     return algorithm->inGrid(pt);
 } 

     return false;
}

/*
 * Check to see if the given coordinate  lies inside the 
 * max value for that axis
 *
 * @param coordinate for a given axis
 * @param max_value for a that axis 
 *
 * @return bool
 */ 
bool BoundaryStrategy::isValid(const int coordinate, const int max_value) {

    return (0 <= coordinate && coordinate < max_value); 
}

/*
 * Set dimensions
 *
 * @param dim Dim3D
 */
void BoundaryStrategy::setDim(const Dim3D theDim) {
//     cerr<<"calling setDim again"<<endl;
    Dim3D oldDim(dim);

    dim = theDim;
    algorithm->setDim(theDim);
    if(! neighborListsInitializedFlag){
      prepareNeighborLists();
      neighborListsInitializedFlag=true;
    }
    
    if(!(oldDim==theDim)){
//       cerr<<"initializeQuickCheckField fcn call"<<endl;
      initializeQuickCheckField(theDim);
    }


}

/*
 * Set current step
 *
 *  @param int currentStep
 */ 
 void BoundaryStrategy::setCurrentStep(const int theCurrentStep) {

   currentStep = theCurrentStep;
   algorithm->setCurrentStep(theCurrentStep);
     
 }

/*
 * Set Irregular
 *
 */
 void BoundaryStrategy::setIrregular() {
  
     regular = false;
     
 }

/*
 * Retrieves a neighbor for a given point after applying
 * boundary conditions as configured
 *
 * @param pt Point3D
 * @param token int
 * @param distance int
 * @param checkBounds bool
 *
 * @return Point3D Valid Neighbor
 */
Point3D BoundaryStrategy::getNeighbor(const Point3D& pt, unsigned int& token,
                                   double& distance, bool checkBounds) {

     Neighbor n;
     Point3D p;
     int x;
     int y;
     int z;
     bool x_bool;
     bool y_bool;
     bool z_bool;

     while(true) {

      // Get a neighbor from the NeighborFinder 
      n = NeighborFinder::getInstance()->getNeighbor(token);
      x = (pt + n.pt).x;
      y = (pt + n.pt).y;
      z = (pt + n.pt).z;

      token++;

      if(!checkBounds || isValid(pt + n.pt)) {
          
         // Valid Neighbor
          break;

      } else {
          
          if (regular) {
              
          // For each coordinate, if it is not valid, apply condition
          x_bool = (isValid(x,dim.x) ? true : strategy_x->applyCondition(x, dim.x));
          y_bool = (isValid(y,dim.y) ? true : strategy_y->applyCondition(y, dim.y));
          z_bool = (isValid(z,dim.z) ? true : strategy_z->applyCondition(z, dim.z));
          
          // If all the coordinates of the neighbor are valid then return the
          // neighbor
            if(x_bool && y_bool && z_bool) {
              
              break; 

             }
            
          } 
          
        }
      

      }
   

     distance = n.distance;
     p.x = x;
     p.y = y;
     p.z = z;
     
     return p;
    
}

int BoundaryStrategy::getNumPixels(int x, int y, int z) const  {

    return algorithm->getNumPixels(x, y, z);

}

//returns true if _offset is stacked in _offsetVec
//false otherwise
bool BoundaryStrategy::checkIfOffsetAlreadyStacked(Point3D & _ptToCheck , std::vector<Point3D> & _offsetVec){

   for(int i = 0 ; i < _offsetVec.size() ; ++i ){
      if( _offsetVec[i].x==_ptToCheck.x && _offsetVec[i].y==_ptToCheck.y && _offsetVec[i].z==_ptToCheck.z)
         return true;
   }
   return false;
}



bool BoundaryStrategy::checkEuclidianDistance(Point3D & _pt1,Point3D & _pt2, float _distance){
   //checks if distance between two points is smaller than _distance
   //used to eliminate in offsetVec offsets that come from periodic conditions (opposite side of the lattice)
   return
   sqrt((double)(_pt1.x-_pt2.x)*(_pt1.x-_pt2.x)+(_pt1.y-_pt2.y)*(_pt1.y-_pt2.y)+(_pt1.z-_pt2.z)*(_pt1.z-_pt2.z))<_distance+0.1;

}

void BoundaryStrategy::prepareNeighborLists(float _maxDistance){
   maxDistance=_maxDistance;
   char a='0';
   
   vector<Point3D> offsetVecTmp;
   vector<float> distanceVecTmp;
//    Dim3D dimTemp(4*maxDistance,4*maxDistance,4*maxDistance);
//    cerr<<"THIS IS DIM"<<dim<<endl;
   Field3DImpl<char> tempField(dim,a);
   
//    float distance=5.0;

   
   //generating list of neighbors
   //We will generate neighbor list by visiting corner points , to account for possible periodic boundary conditions. This should also generate offsets for all the 
   //neighbors also for points in the middle

   vector<Point3D> testPoints;
   //lower plane of the lattice
   testPoints.push_back(Point3D(0,0,0)); 
   testPoints.push_back(Point3D(0,0,dim.z-1));
   testPoints.push_back(Point3D(0,dim.y-1,0));
   testPoints.push_back(Point3D(dim.x-1,0,0));
   //upper plane of the lattice
   testPoints.push_back(Point3D(dim.x-1,dim.y-1,dim.z-1));
   testPoints.push_back(Point3D(0,dim.y-1,dim.z-1));
   testPoints.push_back(Point3D(dim.x-1,0,dim.z-1));
   testPoints.push_back(Point3D(dim.x-1,dim.y-1,0));
   Point3D offset;
   
   
   //This loop generates offsets for neighbors calculations . we eliminate duplicates
   //offsets are stored in offsetVecTmp and distances are stored in distanceVec
   for(int i = 0 ; i < testPoints.size() ; ++i){
      Point3D n;
      Point3D pt=testPoints[i];
      unsigned int token = 0;
      double distance = 0;

      while (true) {
         n = tempField.getNeighbor(pt, token, distance, false);
         if (distance > maxDistance) break;
         
         offset=pt-n;
         if ( !checkIfOffsetAlreadyStacked(offset,offsetVecTmp) && checkEuclidianDistance(pt,n,maxDistance) ){
            offsetVecTmp.push_back(offset);
            distanceVecTmp.push_back(distance);
         }

      }
     

   }


   
   //at this point we have all the offsets for the given simulation but they are unsorted.
   //Sorting  neighbors
   multimap<float,Point3D> sortingMap;
   for( int i = 0 ; i < offsetVecTmp.size() ; ++i){
      sortingMap.insert(make_pair(distanceVecTmp[i],offsetVecTmp[i]));
   }
   //allocating memory for offsetVec and distanceVec
   
   offsetVec.assign(offsetVecTmp.size(),Point3D(0,0,0));
   distanceVec.assign(distanceVecTmp.size(),0.0);
   //Writing content of offsetVec and distanceVec
   int idx=0;
   for (multimap<float,Point3D>::iterator mitr = sortingMap.begin() ; mitr != sortingMap.end() ; ++mitr){
      distanceVec[idx]=mitr->first;
      offsetVec[idx]=mitr->second;
      
      ++idx;
   }
   
   //creating a vector indexed by neighbor order  - entries of this vector are the highest indices of offsets for a 
   //given neighbor order
   float currentDistance=1.0;
   for( int i = 0 ; i < distanceVec.size() ; ++i){
      if(currentDistance<distanceVec[i]){
         neighborOrderIndexVec.push_back(i-i);
         currentDistance=distanceVec[i];
      }
   }
   
    for( int i = 0 ; i < offsetVec.size() ; ++i){
      cerr<<" This is offset["<<i<<"]="<<offsetVec[i]<<" distance="<<distanceVec[i]<<endl;
    }

   
}

bool BoundaryStrategy::isValidDirect(const Point3D &pt) const{
//    cerr<<"xsize="<<checkField.size()<<endl;
//    cerr<<"ysize="<<checkField[0].size()<<endl;
//    cerr<<"zsize="<<checkField[0][0].size()<<endl;

   if(!checkField[pt.x][pt.y][pt.z])
      return true;
   else
      return false;

}
unsigned int BoundaryStrategy::getMaxNeighborIndexFromDepth(float depth){
   //Now determine max neighbor index from a list of neighbor offsets
   unsigned int maxNeighborIndex=0;
   for(int i = 0 ; i < distanceVec.size() ;++i){
      if(distanceVec[i]>depth){
         maxNeighborIndex=i-1;
         break;
      }
   }
   return maxNeighborIndex;
}

Neighbor BoundaryStrategy::getNeighborDirect(Point3D & pt,unsigned int idx, bool checkBounds ) {
   Neighbor n;
   n.pt=pt+offsetVec[idx];

//    if(!isValidDirect(pt)){//here I check whether point, not its neighbor, requires some more checks. If not I return
//       n.distance=distanceVec[idx];
//       return n;
//    }

   //Here I will add condition  if (flagField[pt] ) ...

   if(!checkBounds || isValid(n.pt)) {
          
      // Valid Neighbor
      n.distance=distanceVec[idx];
      return n; 

      } else {
          
          if (regular) {
            bool x_bool;
            bool y_bool;
            bool z_bool;
            int x=n.pt.x;
            int y=n.pt.y;
            int z=n.pt.z;
    
            // For each coordinate, if it is not valid, apply condition
            x_bool = (isValid(x,dim.x) ? true : strategy_x->applyCondition(x, dim.x));
            y_bool = (isValid(y,dim.y) ? true : strategy_y->applyCondition(y, dim.y));
            z_bool = (isValid(z,dim.z) ? true : strategy_z->applyCondition(z, dim.z));
          
            // If all the coordinates of the neighbor are valid then return the
            // neighbor
            if(x_bool && y_bool && z_bool) {
              n.pt.x=x;
              n.pt.y=y;
              n.pt.z=z;
              n.distance= distanceVec[idx];

              return n;

            }else{
               //requesed neighbor does not belong to the lattice
               n.distance=0.0;
               return n;
            }
            
          } 
          
        }


}


void BoundaryStrategy::initializeQuickCheckField(Dim3D _dim){
   //this function will initialize a field of unsigned char with special values:
   //if the value is 0 then no checks whether neighbor belongs to the field will be done
   // other values indicate that further checks are necessary 
   checkField.assign(dim.x, vector<vector< unsigned char > >(dim.y,vector<unsigned char>(dim.z,0)));
   Point3D pt;

   //Now will check initialize 3D array taking into account whether we have 2D or 3D simulation

      for (int x = 0  ; x < dim.x ; ++x)
         for (int y = 0  ; y < dim.y ; ++y)
            for (int z = 0  ; z < dim.z ; ++z){
               //will do case by case initialization - i.e. when field is "flat" in x ,y or z direction
               //Although this is not pretty solution I will leave it for now
               if(x < maxDistance || fabs((float)dim.x-x)<maxDistance ){
                  if(dim.x != 1){//if dim.x is 1 we do not mark this pixel for extra checks
                     checkField[x][y][z]=1;
                  } 
               }
               if(y < maxDistance || fabs((float)dim.y-y)<maxDistance ){
                  if(dim.y != 1){//if dim.y is 1 we do not mark this pixel for extra checks
                     checkField[x][y][z]=1;
                  } 
               }

               if(z < maxDistance || fabs((float)dim.z-z)<maxDistance ){
                  if(dim.z != 1){//if dim.y is 1 we do not mark this pixel for extra checks
                     checkField[x][y][z]=1;
                  } 
               }

            }

}