/*************************************************************************
 *    CompuCell - A software framework for multimodel simulations of     *
 * biocomplexity problems Copyright (C) 2003 University of Notre Dame,   *
 *                             Indiana                                   *
 *                                                                       *
 * This program is free software; IF YOU AGREE TO CITE USE OF CompuCell  *
 *  IN ALL RELATED RESEARCH PUBLICATIONS according to the terms of the   *
 *  CompuCell GNU General Public License RIDER you can redistribute it   *
 * and/or modify it under the terms of the GNU General Public License as *
 *  published by the Free Software Foundation; either version 2 of the   *
 *         License, or (at your option) any later version.               *
 *                                                                       *
 * This program is distributed in the hope that it will be useful, but   *
 *      WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
 *             General Public License for more details.                  *
 *                                                                       *
 *  You should have received a copy of the GNU General Public License    *
 *     along with this program; if not, write to the Free Software       *
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        *
 *************************************************************************/

#include "ConnectivityEnergy.h"

#include <CompuCell3D/Field3D/Field3D.h>
#include <CompuCell3D/Field3D/WatchableField3D.h>
#include <CompuCell3D/Potts3D/Potts3D.h>
#include <CompuCell3D/Potts3D/Cell.h>
#include <CompuCell3D/Automaton/Automaton.h>
using namespace CompuCell3D;

#include <XMLCereal/XMLPullParser.h>
#include <XMLCereal/XMLSerializer.h>

#include <BasicUtils/BasicString.h>
#include <BasicUtils/BasicException.h>

#include <CompuCell3D/Potts3D/Cell.h>

#include <algorithm>

#include <string>
#include <iostream>
using namespace std;

ConnectivityEnergy::ConnectivityEnergy(Potts3D *potts) :
            potts(potts) ,numberOfNeighbors(8)
{
   n.assign(numberOfNeighbors,Point3D(0,0,0)); //allocates memory for vector of neighbors points
   offsetsIndex.assign(numberOfNeighbors,0);//allocates memory for vector of offsetsIndexes
}


void ConnectivityEnergy::addUnique(CellG* cell)
{
   if (!cell) mediumFlag = true;
   //if (!cell) return;  
   // Medium does not count
   for (unsigned int i = 0; i < uniqueCells.size(); i++)
   {
      if (uniqueCells[i] == cell)
          return;
   }
   uniqueCells.push_back(cell);
}

double ConnectivityEnergy::localEnergy(const Point3D &pt) {
  return 0;
}


//modified version of connectivity algorithm - Tinri Aegerter
double ConnectivityEnergy::changeEnergy(const Point3D &pt,
                                  const CellG *newCell,
                                   const CellG *oldCell) {

   if (!oldCell) return 0;//medium doesn't count
   uniqueCells.clear();
   Neighbor neighbor;
   int numOldNeighbors=0;

   if((!newCell)&&(potts->getCellInventory().getCellInventorySize()>1)) return 0;//medium doesn't count

   //rule 0: if oldCell was isolated due to Mitsosis let it die!
   //rule 1: go over first neighrest neighbors and check if newCell has 1st neighrest neioghbotsbelongin to the same cell   
   bool firstFlag=false;
   unsigned int firstMaxIndex=boundaryStrategy->getMaxNeighborIndexFromDepth(1.05);
   for(int i=0 ; i <=firstMaxIndex ; ++i ){
      neighbor=boundaryStrategy->getNeighborDirect(const_cast<Point3D&>(pt),i);
      if(!neighbor.distance){
         //if distance is 0 then the neighbor returned is invalid
         continue;
      }
      if(potts->getCellFieldG()->get(neighbor.pt)==newCell){
         firstFlag=true;
         //break;
      }
       if(potts->getCellFieldG()->get(neighbor.pt)==oldCell){
         numOldNeighbors++;
      }

   }  
   if((numOldNeighbors==0)&&(potts->getCellInventory().getCellInventorySize()>1)){
      return -penalty;
   }
   if(!firstFlag){
      return penalty;
   }


   // Immediate neighbors
      
      for(int i=0 ; i < numberOfNeighbors ; ++i ){
         neighbor=boundaryStrategy->getNeighborDirect(const_cast<Point3D&>(pt),offsetsIndex[i]);
         if(!neighbor.distance){
         //if distance is 0 then the neighbor returned is invalid
         continue;
         }

         n[i]=neighbor.pt;
      }

   //rule 2: Count collisions between pixels belonging to different cells. If the collision count is greater than 2 reject the spin flip
   mediumFlag = false;
   int collisioncount = 0;
   for (unsigned int i = 0; i < numberOfNeighbors; i++)
   {
      if (
            ((potts->getCellFieldG()->get(n[i]) == oldCell) || (potts->getCellFieldG()->get(n[(i+1) % numberOfNeighbors]) == oldCell))
              &&
            (potts->getCellFieldG()->get(n[i]) !=potts->getCellFieldG()->get(n[(i+1) % numberOfNeighbors]))
         )
      {
        addUnique(potts->getCellFieldG()->get(n[i]));
        addUnique(potts->getCellFieldG()->get(n[(i+1)%numberOfNeighbors]));
	collisioncount++;
      }
   }

   if (collisioncount == 2) return 0;  // Accept
   else { // Conditional rejection

      return penalty;  // Reject
   }
 
}

// Original Implementation - kept as a reference
// double ConnectivityEnergy::changeEnergy(const Point3D &pt,
//                                   const CellG *newCell,
//                                    const CellG *oldCell) {
//  
//    if (!oldCell) return 0;
//    uniqueCells.clear();
// 
//    // Immediate neighbors
//       Neighbor neighbor;
//       for(int i=0 ; i < numberOfNeighbors ; ++i ){
//          neighbor=boundaryStrategy->getNeighborDirect(const_cast<Point3D&>(pt),offsetsIndex[i]);
//          if(!neighbor.distance){
//          //if distance is 0 then the neighbor returned is invalid
//          continue;
//          }
// 
//          n[i]=neighbor.pt;
//       }
// 
// 
// 
// 
//    mediumFlag = false;
//    int collisioncount = 0;
//    for (unsigned int i = 0; i < numberOfNeighbors; i++)
//    {
//       if (((potts->getCellFieldG()->get(n[i]) == oldCell) ||
//            (potts->getCellFieldG()->get(n[(i+1) % numberOfNeighbors]) == oldCell))
//               &&
//           (potts->getCellFieldG()->get(n[i]) !=
//            potts->getCellFieldG()->get(n[(i+1) % numberOfNeighbors])))
//       {
//         addUnique(potts->getCellFieldG()->get(n[i]));
//         addUnique(potts->getCellFieldG()->get(n[(i+1)%numberOfNeighbors]));
// 	collisioncount++;
//       }
//    }
//    //if (collisioncount == 0) {
//    //    cout << "COLLISION COUNT OF 0" << endl;
//        /*cout << potts->getCellFieldG()->get(n[0]) << endl;
//        cout << potts->getCellFieldG()->get(n[1]) << endl;
//        cout << potts->getCellFieldG()->get(n[2]) << endl;
//        cout << potts->getCellFieldG()->get(n[3]) << endl;
//        cout << potts->getCellFieldG()->get(n[4]) << endl;
//        cout << potts->getCellFieldG()->get(n[5]) << endl;
//        cout << potts->getCellFieldG()->get(n[6]) << endl;
//        cout << potts->getCellFieldG()->get(n[7]) << endl;
//        int aaa;
//        cin >> aaa;*/
//    //}
//    if (collisioncount == 2) return 0;  // Accept
//    else { // Conditional rejection
//       if (collisioncount > 2 && uniqueCells.size() == 2 && !mediumFlag) return 0; // Accept
//       else return penalty;  // Reject
//    }
// }




void ConnectivityEnergy::readXML(XMLPullParser &in) {
  in.skip(TEXT);

  while (in.check(START_ELEMENT)) {
    if (in.getName() == "Penalty") {
      penalty = BasicString::parseDouble(in.matchSimple());
      
    }
   else {
      throw BasicException(string("Unexpected element '") + in.getName() +
                           "'!", in.getLocation());
    }

    in.skip(TEXT);
  }
   
   //here we initialize offsets for neighbors (up to second nearest) in 2D

   initializeNeighborsOffsets();

}

void ConnectivityEnergy::orderNeighborsClockwise
(Point3D & _midPoint, const std::vector<Point3D> & _offsets, std::vector<int> & _offsetsIndex)
{
      //this function maps indexes of neighbors as given by boundary strategy to indexes of neighbors which are in clockwise order
      //that is when iterating over list of neighbors using boundaryStrategy API you can retrieve neighbors in clocwise order if you use
      //neighbor=boundaryStrategy->getNeighborDirect(const_cast<Point3D&>(pt),offsetsIndex[i]) call
      Neighbor neighbor;
      for(unsigned int nIdx=0 ; nIdx <= maxNeighborIndex ; ++nIdx ){
         neighbor=boundaryStrategy->getNeighborDirect(const_cast<Point3D&>(_midPoint),nIdx);
         if(!neighbor.distance){
         //if distance is 0 then the neighbor returned is invalid
         continue;
         }
         for(int i = 0 ; i < numberOfNeighbors ; ++i){
            if(_midPoint+_offsets[i] == neighbor.pt){
               _offsetsIndex[i]=nIdx;
            }
         }
      }
}


void ConnectivityEnergy::initializeNeighborsOffsets(){
   
   Dim3D fieldDim=potts->getCellFieldG()->getDim();
   vector<Point3D> offsets;

   offsets.assign(numberOfNeighbors,Point3D(0,0,0));//allocates memory for vector of offsets 

   boundaryStrategy=BoundaryStrategy::getInstance();
   maxNeighborIndex=0;

   maxNeighborIndex=boundaryStrategy->getMaxNeighborIndexFromDepth(1.45);//second nearest neighbor

   ASSERT_OR_THROW("This plugin will only work for 2D simulations i.e. one lattice dimansion must be equal to 1 Your simulations appears to be 3D", !(fieldDim.x>1 && fieldDim.y>1 && fieldDim.z>1 ) );

   //here we define neighbors  offsets in the "clockwise order"
   if(fieldDim.x==1){
      //clockwise offsets
      offsets[0]=Point3D(0,0,-1);
      offsets[1]=Point3D(0,-1,-1);
      offsets[2]=Point3D(0,-1,0);
      offsets[3]=Point3D(0,-1,1);
      offsets[4]=Point3D(0,0,1);
      offsets[5]=Point3D(0,1,1);
      offsets[6]=Point3D(0,1,0);
      offsets[7]=Point3D(0,1,-1);

      Point3D midPoint(0, fieldDim.y/2, fieldDim.z/2);
      orderNeighborsClockwise(midPoint , offsets , offsetsIndex);

   }

   if(fieldDim.y==1){

      offsets[0]=Point3D(0,0,-1);
      offsets[1]=Point3D(-1,0,-1);
      offsets[2]=Point3D(-1,0,0);
      offsets[3]=Point3D(-1,0,1);
      offsets[4]=Point3D(0,0,1);
      offsets[5]=Point3D(1,0,1);
      offsets[6]=Point3D(1,0,0);
      offsets[7]=Point3D(1,0,-1);

      Point3D midPoint(fieldDim.x/2, 0, fieldDim.z/2);
      orderNeighborsClockwise(midPoint , offsets , offsetsIndex);



   }
   
   if(fieldDim.z==1){
      offsets[0]=Point3D(0,-1,0);
      offsets[1]=Point3D(-1,-1,0);
      offsets[2]=Point3D(-1,0,0);
      offsets[3]=Point3D(-1,1,0);
      offsets[4]=Point3D(0,1,0);
      offsets[5]=Point3D(1,1,0);
      offsets[6]=Point3D(1,0,0);
      offsets[7]=Point3D(1,-1,0);
      
      Point3D midPoint(fieldDim.x/2, fieldDim.y/2, 0);
      orderNeighborsClockwise(midPoint , offsets , offsetsIndex);
   }

}


void ConnectivityEnergy::writeXML(XMLSerializer &out) {
}
