/*************************************************************************
 *    CompuCell - A software framework for multimodel simulations of     *
 * biocomplexity problems Copyright (C) 2003 University of Notre Dame,   *
 *                             Indiana                                   *
 *                                                                       *
 * This program is free software; IF YOU AGREE TO CITE USE OF CompuCell  *
 *  IN ALL RELATED RESEARCH PUBLICATIONS according to the terms of the   *
 *  CompuCell GNU General Public License RIDER you can redistribute it   *
 * and/or modify it under the terms of the GNU General Public License as *
 *  published by the Free Software Foundation; either version 2 of the   *
 *         License, or (at your option) any later version.               *
 *                                                                       *
 * This program is distributed in the hope that it will be useful, but   *
 *      WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
 *             General Public License for more details.                  *
 *                                                                       *
 *  You should have received a copy of the GNU General Public License    *
 *     along with this program; if not, write to the Free Software       *
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        *
 *************************************************************************/

#include "MitosisPlugin.h"

#include <CompuCell3D/Simulator.h>
#include <CompuCell3D/Field3D/Field3D.h>
#include <CompuCell3D/Potts3D/Potts3D.h>
#include <CompuCell3D/Potts3D/Cell.h>
#include <CompuCell3D/Automaton/Automaton.h>
using namespace CompuCell3D;

#include <XMLCereal/XMLPullParser.h>
#include <XMLCereal/XMLSerializer.h>

#include <BasicUtils/BasicString.h>
#include <BasicUtils/BasicException.h>

#include <iostream>
using namespace std;

MitosisPlugin::MitosisPlugin() {potts=0; split=false;}

MitosisPlugin::~MitosisPlugin() {}

void MitosisPlugin::init(Simulator *simulator) {

  potts = simulator->getPotts();


   simulator->getPotts()->registerCellGChangeWatcher(this);
   simulator->getPotts()->registerStepper(this);
   turnOn();
}


void MitosisPlugin::field3DChange(const Point3D &pt, CellG *newCell,
				  CellG *oldCell) {

  // Note: We cannot mitosis the cell here because  updating the field
  // from this function would cause further field3DChange calls which
  // would be out of order for any change listeners who  happened to
  // be after this one in the list.


  if (newCell){
//       cerr<<"this is the mitosis newCell "<<newCell<<endl;
//       cerr<<" DoublingVolume="<<doublingVolume<<endl;
      if(newCell->volume>= doublingVolume){
         split = true;
         splitPt = pt;

      }
  }

}

void MitosisPlugin::turnOn() {on = true;}
void MitosisPlugin::turnOff() {on = false;}

void MitosisPlugin::step() {
   bool didMitosis=doMitosis();
   if(didMitosis){
      updateAttributes();
    
   }
   
}

CellG * MitosisPlugin::getChildCell(){return childCell;}
CellG * MitosisPlugin::getParentCell(){return parentCell;}


void MitosisPlugin::updateAttributes(){

   ///copying type and target volume of the parent cell to the new cell
   childCell->type = parentCell->type;
   childCell->targetVolume = parentCell->targetVolume;
}

bool MitosisPlugin::doMitosis(){
   
  bool didMitosis=false;

  if (split && on) {
    split = false;
    

    Field3D<CellG *> *cellField = potts->getCellFieldG();
    //reseting poiters to parent and child cell - neessary otherwise may get some strange side effects when mitisis is aborted
    childCell=0;
    parentCell=0;

    CellG *cell = cellField->get(splitPt);
    parentCell=cell;

    ASSERT_OR_THROW("Cell should not be NULL at mitosis point!", cell);

    int volume = cell->volume;
    int newVol = 0;
    int targetVol = volume / 2;



    vector<Point3D> ary0Vec;
    vector<Point3D> ary1Vec;

    vector<Point3D> *tmp;;
    vector<Point3D> *ary0=&ary0Vec;
    vector<Point3D> *ary1=&ary1Vec;
    
    ary0->clear();
    ary1->clear();

    // Put the first point in the array
    ary0->push_back(splitPt);

    CellG *splitCell = 0;

    // Do a breadth first search for approximately half of the  cell's
    // volume.
    
    while (ary0->size() > 0 && newVol < targetVol) {
      Point3D n;
      CellG *nCell;
      // Loop over all the points from the last round.
      for (unsigned int i = 0; i < ary0->size(); i++) {
         unsigned int token = 0;
         double distance = 0;

         while (distance <= 3.0) {
         n = cellField->getNeighbor((*ary0)[i], token, distance);

      
         // Add to split cell
          nCell=cellField->get(n);

         
         // Add to split cell
         if (nCell/*cellField->get(n)*/ == cell) {

            ary1->push_back(n);

            newVol++;
            if (splitCell){
               cellField->set(n, splitCell);
            }
            else{
               splitCell = potts->createCellG(n);
               childCell=splitCell;
            }
            if (newVol >= targetVol) break;
         }
         }

         if (newVol >= targetVol) break;
      }

      // Swap arrays;
      tmp = ary0;
      ary0 = ary1;
      ary1 = tmp;
      ary1->clear();
    }
   if(!childCell){
      cerr<<"Fragmented Cell - mitosis aborted"<<endl;
      didMitosis=false;
      return didMitosis;
   }
   
   if(childCell && fabs((float)childCell->volume-parentCell->volume)>2.0){
      cerr<<"cell was fragmented before mitosis, volumes of parent and child cells might significantly differ"<<endl;
//       cerr<<"C++ childCell.volume="<<childCell->volume<<" parentCell.volume="<<parentCell->volume<<endl;
      didMitosis=true;
      return didMitosis;
   }

  }
didMitosis=true;
return didMitosis;


}
void MitosisPlugin::readXML(XMLPullParser &in) {
  in.skip(TEXT);

  while (in.check(START_ELEMENT)) {
    if (in.getName() == "DoublingVolume") {
      doublingVolume = BasicString::parseUInteger(in.matchSimple());
      ASSERT_OR_THROW("DoublingVolume must be > 1!", doublingVolume > 1);
      
    } else {
      THROW(string("Unexpected element '") + in.getName() + "'!");
    }
    
    in.skip(TEXT);
  }
}

void MitosisPlugin::writeXML(XMLSerializer &out) {
}



















