/*************************************************************************
 *    CompuCell - A software framework for multimodel simulations of     *
 * biocomplexity problems Copyright (C) 2003 University of Notre Dame,   *
 *                             Indiana                                   *
 *                                                                       *
 * This program is free software; IF YOU AGREE TO CITE USE OF CompuCell  *
 *  IN ALL RELATED RESEARCH PUBLICATIONS according to the terms of the   *
 *  CompuCell GNU General Public License RIDER you can redistribute it   *
 * and/or modify it under the terms of the GNU General Public License as *
 *  published by the Free Software Foundation; either version 2 of the   *
 *         License, or (at your option) any later version.               *
 *                                                                       *
 * This program is distributed in the hope that it will be useful, but   *
 *      WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
 *             General Public License for more details.                  *
 *                                                                       *
 *  You should have received a copy of the GNU General Public License    *
 *     along with this program; if not, write to the Free Software       *
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        *
 *************************************************************************/

#ifndef MITOSISPLUGIN_H
#define MITOSISPLUGIN_H

#include <CompuCell3D/Plugin.h>

#include <CompuCell3D/Potts3D/Cell.h>

#include <CompuCell3D/dllDeclarationSpecifier.h>
#include <CompuCell3D/Potts3D/EnergyFunction.h>
#include <XMLCereal/XMLSerializable.h>
#include <CompuCell3D/Field3D/Dim3D.h>
#include <CompuCell3D/Potts3D/CellGChangeWatcher.h>
#include <CompuCell3D/Potts3D/Stepper.h>

#include <CompuCell3D/Steppable.h>
#include <CompuCell3D/Potts3D/Cell.h>
#include <CompuCell3D/Potts3D/CellGChangeWatcher.h>
#include <BasicUtils/BasicClassAccessor.h>
#include <BasicUtils/BasicClassGroup.h> //had to include it to avoid problems with template instantiation
#include <CompuCell3D/Field3D/AdjacentNeighbor.h>
#include <CompuCell3D/dllDeclarationSpecifier.h>

#include <map>
#include <vector>
#include <set>

template <class T>
class BasicClassAccessor;

#include <BasicUtils/BasicArray.h>
#include <CompuCell3D/steppables/GRNcells/GRNconcentrationsData.h>
#include <CompuCell3D/steppables/GRNcells/GRNcells.h>

namespace CompuCell3D {
  class Potts3D;
  class Simulator;
  class Dim3D;
  class BoundaryStrategy;
  class Potts3D;
  class CellTypePlugin;
  class Cell;
  class Field3DIndex;
  template <class T> class Field3D;
  template <class T> class WatchableField3D;
  class CellInventory;


  class PointDist{

	  public: Point3D P;
			  float dist;
	  PointDist(Point3D P, float dist): P(P), dist(dist) {}
	  static bool cmp( PointDist a, PointDist b ){
		  return a.dist > b.dist;
	  }

  };

//some of the geometry code from / inspired by softsurfer.com
//    Plane with a point and a normal {Point V0; Vector n;}
  class Plane{
  public: Point3D V0; Point3D n;
  
    Plane(Point3D V0, Point3D n) :
      V0(V0), n(n) {}

// dot product (3D) which allows vector operations in arguments
#define dot(u,v)   ((u).x * (v).x + (u).y * (v).y + (u).z * (v).z)
#define norm(v)    sqrt((float)dot(v,v))  // norm = length of vector
#define d(u,v)     norm(u-v)       // distance = norm of difference
#define cross(u,v) Point3D( ( (u).y*(v).z-(u).z*(v).y ) , ( (u).z*(v).x-(u).x*(v).z ) , ( (u).x*(v).y-(u).y*(v).x )  )
#define PI 3.14159265

	  //construct from triangle
	  Plane(Point3D V0, Point3D V1, Point3D V2) : 
	  V0(V0)    {
		Point3D nl=cross((V1-V0),(V2-V0));
		float nn=norm(nl);
		nl.x=nl.x/nn;
		nl.y=nl.y/nn;
		nl.z=nl.z/nn;

		this->n=nl;
	
	}


// pbase_Plane(): get base of perpendicular from point to a plane
//    Input:  P = a 3D point
//            PL = a plane with point V0 and normal n
//    Output: *B = base point on PL of perpendicular from P
//    Return: the distance from P to the plane PL
float pbase_Plane( Point3D P, Plane PL, Point3D* B)
{
    float    sb, sn, sd;

    sn = -dot( PL.n, (P - PL.V0));
    sd = dot(PL.n, PL.n);
    sb = sn / sd;

	//*B = P + sb * PL.n;
	Point3D ttt=P;
	ttt.x+=PL.n.x*sb;
	ttt.y+=PL.n.y*sb;
	ttt.z+=PL.n.z*sb;
    *B = ttt ;
    return d(P, *B);
}
  };//end Plane class

  class MitosisPlugin : public Plugin, public CellGChangeWatcher,
			public Stepper {
    protected:     
    Potts3D *potts;
    
    unsigned int doublingVolume;

    CellG *childCell;
    CellG *parentCell;
    Field3D<CellG *> *cellField;
    
    Point3D splitPt;
    bool split;
    bool on;
    
    BasicArray<Point3D> arrays[2];

	BasicClassAccessor<GRNconcentrationsData> * GRNconcentrationsDataAccessorPtr;


	bool MitosisPlugin::lonely(Point3D &m, Point3D &n);

  public:
    MitosisPlugin();
    virtual ~MitosisPlugin();

    // SimObject interface
    virtual void init(Simulator *simulator);

    // CellChangeWatcher interface
    virtual void field3DChange(const Point3D &pt, CellG *newCell,
                               CellG *oldCell);


    unsigned int getDoublingVolume(){return doublingVolume;}
    void setDoublingVolume(unsigned int _doublingVolume){doublingVolume=_doublingVolume;}
    // Stepper interface
    virtual void step();

    // Functions to turn on and off
    virtual void turnOn();
    virtual void turnOff();
    virtual bool doMitosis();///actually make mitosis - returns true if mitosis was done , false otherwise
    virtual void updateAttributes();///updates some of the cell attributes 
    CellG * getChildCell();
    CellG * getParentCell();

    void setPotts(Potts3D *_potts){potts=_potts;}    
    // Begin XMLSerializable interface
    virtual void readXML(XMLPullParser &in);
    virtual void writeXML(XMLSerializer &out);
    // End XMLSerializable interface
  };
};
#endif
