/*************************************************************************
 *    CompuCell - A software framework for multimodel simulations of     *
 * biocomplexity problems Copyright (C) 2003 University of Notre Dame,   *
 *                             Indiana                                   *
 *                                                                       *
 *   THIS PLUGIN (C) BY JOHANNES KNABE, UNIVERSITY OF HERTFORDSHIRE,UK   *
 *                                                                       *
 * This program is free software; IF YOU AGREE TO CITE USE OF CompuCell  *
 *  IN ALL RELATED RESEARCH PUBLICATIONS according to the terms of the   *
 *  CompuCell GNU General Public License RIDER you can redistribute it   *
 * and/or modify it under the terms of the GNU General Public License as *
 *  published by the Free Software Foundation; either version 2 of the   *
 *         License, or (at your option) any later version.               *
 *                                                                       *
 * This program is distributed in the hope that it will be useful, but   *
 *      WITHOUT ANY WARRANTY; without even the implied warranty of       *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU    *
 *             General Public License for more details.                  *
 *                                                                       *
 *  You should have received a copy of the GNU General Public License    *
 *     along with this program; if not, write to the Free Software       *
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.        *
 *************************************************************************/

#include <CompuCell3D/Automaton/Automaton.h>
#include <CompuCell3D/Simulator.h>
#include <CompuCell3D/Potts3D/Cell.h>
#include <CompuCell3D/Potts3D/Potts3D.h>
#include <CompuCell3D/Field3D/Point3D.h>
#include <CompuCell3D/Field3D/Dim3D.h>
#include <CompuCell3D/plugins/CellType/CellTypePlugin.h>
#include <CompuCell3D/plugins/ContactMultiCad/ContactMultiCadEnergy.h>
#include <CompuCell3D/Potts3D/CellInventory.h>
#include <CompuCell3D/Field3D/Field3D.h>
#include <CompuCell3D/Field3D/WatchableField3D.h>
#include <BasicUtils/BasicClassAccessor.h>
#include <CompuCell3D/plugins/ContactMultiCad/ContactMultiCadData.h>

using namespace CompuCell3D;

//define EXP_STL
#include <BasicUtils/BasicString.h>
#include <BasicUtils/BasicException.h>
#include "GRNcells.h"

#include <XMLCereal/XMLPullParser.h>
#include <XMLCereal/XMLSerializer.h>

#include <CompuCell3D/Potts3D/CellInventory.h>
#include "./KnabeGRN.h"
#include "./GRNconcentrationsData.h"

#include <CompuCell3D/Potts3D/Cell.h>
#include <PublicUtilities/StringUtils.h>

#include <string>
#include <sstream>
#include <fstream>
#include <map>
#include <cmath>
#include <iostream>
#include <algorithm>

using namespace std;

BasicClassAccessor<GRNconcentrationsData> GRNconcentrationsDataAccessor;

GRNcells::GRNcells() :
  potts(0)
{}

void GRNcells::init(Simulator *simulator) {
this->simulator=simulator;
potts = simulator->getPotts();
setGRNconcentrationsDataAccessorPtr(&GRNconcentrationsDataAccessor);
potts->getCellFactoryGroupPtr()->registerClass(&GRNconcentrationsDataAccessor);

}

void GRNcells::start() {}

/*
Given a cell searches for one point on the CA grid that is part of it.
A good starting point should be passed to decrease search time, for example
the center of the cell's mass! [NO SEARCH YET, ONLY CHECKS NEIGHBORS!
IN THE FUTURE THIS WILL RECURSIVELY INCREASE RANGE (1.75)]
*/
bool GRNcells::findCellPoint(CellG* cell,Point3D &cm){
	void* cF=potts->getCellFieldG();
    WatchableField3D<CellG *> *fieldG=(WatchableField3D<CellG *> *)cF;
	CellG * nCell; Neighbor neighbor;
	nCell = fieldG->get(cm);
	if((nCell)&&(nCell->id==cell->id))
		return true;
	BoundaryStrategy *boundaryStrategy=BoundaryStrategy::getInstance();
	unsigned int maxNeighborIndex=boundaryStrategy->getMaxNeighborIndexFromDepth(1.75);
	cerr<"CA neighbors: ";
	for(unsigned int nIdx=0 ; nIdx <= maxNeighborIndex ; ++nIdx ){
		neighbor=boundaryStrategy->getNeighborDirect(cm,nIdx);
		if(!neighbor.distance){
			//if distance is 0 then the neighbor returned is invalid
				continue;
		}
		nCell = fieldG->get(neighbor.pt);
		if((nCell)&&(nCell->id==cell->id)){
			cm.x=neighbor.pt.x;
			cm.y=neighbor.pt.y;
			cm.z=neighbor.pt.z;
			return true;
		}
		cerr<<neighbor.pt.x<<"/"<<neighbor.pt.y<<", ";
	}//end for
	cerr<<endl;
	return false;
}

void GRNcells::getCellPixels(CellG * cell,Point3D &cm,vector<Point3D>& cP)
{			CellG * nCell; Neighbor neighbor;
			void* cF=potts->getCellFieldG();
			WatchableField3D<CellG *> *fieldG=(WatchableField3D<CellG *> *)cF;
			BoundaryStrategy *boundaryStrategy=BoundaryStrategy::getInstance();
			unsigned int maxNeighborIndex=boundaryStrategy->getMaxNeighborIndexFromDepth(1.05);
			for(unsigned int nIdx=0 ; nIdx <= maxNeighborIndex ; ++nIdx ){
				neighbor=boundaryStrategy->getNeighborDirect(cm,nIdx);
				if(!neighbor.distance){
					//if distance is 0 then the neighbor returned is invalid
					continue;
				}
				nCell = fieldG->get(neighbor.pt);
				if((nCell)&&(nCell->id==cell->id)){
					//if not in list yet{
					bool inList=false;
					for(int i=0;i<cP.size();i++){
						if((cP.at(i).x==neighbor.pt.x)&&(cP.at(i).y==neighbor.pt.y)&&(cP.at(i).z==neighbor.pt.z)){
							inList=true;
							break;
						}
					}
					if(!inList){
						//cerr<<"Pushing neighbor: "<<neighbor.pt.x<<"/"<<neighbor.pt.y<<" "<<endl;
						Point3D cPpush(neighbor.pt.x,neighbor.pt.y,neighbor.pt.z);
						cP.push_back(cPpush);
						getCellPixels(cell,neighbor.pt,cP);
					}
				}
			}//end for
}

//absolute: set all pixel concentrations to avg value
void GRNcells::setAvgConcentration(vector<Point3D>& pixels, std::string field, float avgConc){
			std::map<std::string,CompuCell3D::Field3DImpl<float>*> & _fieldMap=simulator->getConcentrationFieldNameMap();
			std::map<std::string,Field3DImpl<float>*>::iterator mitr;
			for( mitr = _fieldMap.begin(); mitr != _fieldMap.end() ; ++mitr)
				if(mitr->first.compare(field)==0){
					for(int i=0;i<pixels.size();i++){
						mitr->second->set(pixels.at(i),avgConc);
					}//end for
					break;
				}//end if
}

//relative: every pixels value gets changed by changeConc. 
//N.B. if changeConc<0: negative values are not allowed.
void GRNcells::changeConcentration(vector<Point3D>& pixels, std::string field, float changeConc){
			std::map<std::string,CompuCell3D::Field3DImpl<float>*> & _fieldMap=simulator->getConcentrationFieldNameMap();
			std::map<std::string,Field3DImpl<float>*>::iterator mitr;
			float change;
			for( mitr = _fieldMap.begin(); mitr != _fieldMap.end() ; ++mitr)
				if(mitr->first.compare(field)==0){
					for(int i=0;i<pixels.size();i++){
						change=mitr->second->get(pixels.at(i))+changeConc;
						if(change<0.0) change=0.0;
						mitr->second->set(pixels.at(i),change);
					}//end for
					break;
				}//end if
}

//returns pixels' average concentration for chemical field, or a negative value if something goes wrong
float GRNcells::getAvgConcentration(vector<Point3D>& pixels, std::string field){
			//cerr<<"Cell "<<cell->id<<" has "<<cP.size()<<" pixels according to us... Volume: "<<cell->volume<<endl;
			float sum=0.0;
			std::map<std::string,CompuCell3D::Field3DImpl<float>*> & _fieldMap=simulator->getConcentrationFieldNameMap();
			std::map<std::string,Field3DImpl<float>*>::iterator mitr;
			for( mitr = _fieldMap.begin(); mitr != _fieldMap.end() ; ++mitr)
				if(mitr->first.compare(field)==0){
					for(int i=0;i<pixels.size();i++){
						sum+=mitr->second->get(pixels.at(i));
					}//end for
					break;
				}//end if
			return sum/pixels.size();
}

void GRNcells::step(const unsigned int currentStep) {
CellInventory *cellInventoryPtr=& potts->getCellInventory();

CellG * cell;
unsigned int count=0;
CellInventory::cellInventoryIterator cInvItr;
for( cInvItr=cellInventoryPtr->cellInventoryBegin();
	 cInvItr!=cellInventoryPtr->cellInventoryEnd();
	 ++cInvItr){
		 count++;
		 cell=*cInvItr;
		 if(cell->id!=3){
			if((currentStep>30)&&(cellInventoryPtr->getCellInventorySize()<30))
			   cell->targetVolume=25;//FOR MITSOSIS
			 else
			   cell->targetVolume=15;
			 cell->lambdaVolume=20;
			 cell->targetSurface=1;
			 cell->lambdaSurface=1;
		 }

		 ContactMultiCadPlugin * ttt=(ContactMultiCadPlugin *) simulator->pluginManager.get("ContactMultiCad");
		 if(ttt){
			BasicClassAccessor<ContactMultiCadData> * contactMultiCadDataAccessor = ttt->getContactMultiCadDataAccessorPtr();
			if(cell->id==1){
			  contactMultiCadDataAccessor->get(cell->extraAttribPtr)->assignValue(0,1.0);
			  GRNconcentrationsDataAccessorPtr->get(cell->extraAttribPtr)->assignValue(0,0.77);
			  cerr<<"Concentration value is: "<<GRNconcentrationsDataAccessorPtr->get(cell->extraAttribPtr)->getValue(0)<<endl;
  			  cell->type=2;
			}else{
				if(GRNconcentrationsDataAccessorPtr->get(cell->extraAttribPtr)->getValue(0)>0.1){
					cerr<<"copying works!"<<endl;
					exit(0);
				}
			}
			if(cell->id==2){
			  contactMultiCadDataAccessor->get(cell->extraAttribPtr)->assignValue(1,1.0);
			  contactMultiCadDataAccessor->get(cell->extraAttribPtr)->assignValue(2,1.0);
  			  cell->type=2;
			}
			if(cell->id==3){
				if(currentStep==300){
				  cell->targetVolume=0;
				  cell->targetSurface=0;
				  cell->type=1;
				}
			  contactMultiCadDataAccessor->get(cell->extraAttribPtr)->assignValue(2,1.0);
  			  //cell->type=2;
			  cerr<<"cell no. 3's volume: "<<cell->volume<<" target: "<<cell->targetVolume<<endl;
			}
		 }


		 //cout<<"ID: "<<cell->id<<" Target-Vol: "<<cell->targetVolume<<" Real-Vol: "<<cell->volume<<" Target-Sur: "<<cell->targetSurface<<" Real-Sur: "<<cell->surface<<endl;
		 //cout<<"ID: "<<cell->id<<" Mass center: "<<(cell->xCM/cell->volume)<<" / "<<(cell->yCM/cell->volume)<<endl;
		 if(cell->id==1){
			Point3D cm(cell->xCM/cell->volume,cell->yCM/cell->volume,cell->zCM/cell->volume);
			if(findCellPoint(cell,cm)){
				cout<<"Yeah, starting point found!"<<endl;
				vector<Point3D> pixels(0);
				pixels.push_back(cm);
				getCellPixels(cell,cm,pixels);
				if(currentStep==50)
					changeConcentration(pixels,"SP1",0.5);
				cout<<"Cell 1 has avg. concentrations - SP1: "<<getAvgConcentration(pixels,"SP1")<<" SP2: "<<getAvgConcentration(pixels,"SP2")<<endl;;
			}
		 }
}//end for cellIterator


/*
   CellInventory *cellInventoryPtr=& potts->getCellInventory();
   CellInventory::cellInventoryIterator cInvItr;
   CellG * cell;
   ///loop over all the cells in the inventory   
   for( cInvItr = cellInventoryPtr->cellInventoryBegin() ; cInvItr !=cellInventoryPtr->cellInventoryEnd() ; ++cInvItr ){
      cell=*cInvItr;
         if(cell->id==id)
            break;
   }
	if(cell->id==id){

//use center of mass as starting point for search 
//cout<<"ID: "<<cell->id<<" Mass center: "<<(cell->xCM/cell->volume)<<" / "<<(cell->yCM/cell->volume)<<endl;
		Point3D cm(cell->xCM/cell->volume,cell->yCM/cell->volume,cell->zCM/cell->volume);
		if(findCellPoint(cell,cm)){
			cout<<"Yeah, starting point found!"<<endl;
			vector<Point3D> cP(0);
			cP.push_back(cm);
			getCellPixels(cell,cm,cP);

		}

	}
	cerr<<"ERROR: CELL "<<id<<" NOT FOUND!"<<endl;
	exit(1);
	return -1.0;

*/
if(currentStep==10){//DIFFUSION EXAMPLE

   std::map<std::string,CompuCell3D::Field3DImpl<float>*> & _fieldMap=simulator->getConcentrationFieldNameMap();
   std::map<std::string,Field3DImpl<float>*>::iterator mitr;
   for( mitr = _fieldMap.begin(); mitr != _fieldMap.end() ; ++mitr){
      cerr<<"GOT FIELD: "<<mitr->first<<endl;

/*	  if(mitr->first.compare("SP1")==0){
		CompuCell3D::Dim3D fieldDim=mitr->second->getDim();
		cerr<<"Dimension:"<<fieldDim<<" xxx "<<mitr->first.c_str()<<endl;
		Point3D x(0,0,0);
		mitr->second->set(x,1.0);
	  }*/
	  if(mitr->first.compare("SP2")==0){
		CompuCell3D::Dim3D fieldDim=mitr->second->getDim();
		cerr<<"Dimension:"<<fieldDim<<" xxx "<<mitr->first.c_str()<<endl;
		Point3D x(10,10,0);
		mitr->second->set(x,1.0);
	  }
   }
}

cout<<"Number of cells: "<<count<<endl;
		// Sleep(500);
}

void GRNcells::finish() {}

void GRNcells::readXML(XMLPullParser &in) {
 in.skip(TEXT);
 while (in.check(START_ELEMENT)) {
        if (in.getName() == "Genome") {
      
		   std::string geno=in.matchSimple();
     	   KnabeGRN *t=new KnabeGRN(geno);
		   t->process(0.5);
		   t->process(0.5);
		   t->process(0.5);
		   t->process(0.5);
		   t->process(0.5);
		   t->process(0.5);
		   t->process(0.5);
		   t->process(0.5);
		   t->process(0.5);
		   t->process(0.5);
		   t->process(0.5);
		   t->process(0.5);
		   t->process(0.5);
		   for(int i=0;i<t->genes.size();i++)
			   if((t->genes[i]!=NULL)&&(t->genes[i]->proteinSiteActivations.size()>0)){
				   cout<<"g"<<i<<": "<<endl;
				   cout<<t->genes[i]->proteinSiteActivations[0][0]<<endl;
			   }
		 cout<<"cisFor0: "<<t->genes[0]->getNoOfCisSitesForProt(0)<<endl;
		 cout<<"cisFor1: "<<t->genes[0]->getNoOfCisSitesForProt(1)<<endl;
		 cout<<"cisFor2: "<<t->genes[0]->getNoOfCisSitesForProt(2)<<endl;
		 cout<<"cisFor3: "<<t->genes[0]->getNoOfCisSitesForProt(3)<<endl;
		 cout<<"cisFor4: "<<t->genes[0]->getNoOfCisSitesForProt(4)<<endl;
		 cout<<"cisFor5: "<<t->genes[0]->getNoOfCisSitesForProt(5)<<endl;
		 cout<<"cisFor6: "<<t->genes[0]->getNoOfCisSitesForProt(6)<<endl;
		 cout<<"cisFor7: "<<t->genes[0]->getNoOfCisSitesForProt(6)<<endl;
         //checkFreq = BasicString::parseUInteger();
         //checkSanity=true;
   
      }
      else {
         throw BasicException(string("Unexpected element '") + in.getName() +
                              "'!", in.getLocation());
      }
   
      in.skip(TEXT);
   }
}

void GRNcells::writeXML(XMLSerializer &out) {}
